@@ -27,7 +27,6 @@
  * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  * POSSIBILITY OF SUCH DAMAGE.
  */
-/* $Id: obj_RegExp.c 1282 2007-09-13 22:26:35Z d $ */
 
 #if HAVE_CONFIG_H
 # include <config.h>
@@ -51,7 +50,7 @@
 #include "stringdefs.h"
 #include "init.h"
 #include "nmath.h"
-#include "parse.h"
+#include "compare.h"
 
 /*
  * 15.10 The RegExp object.
