@@ -1,5 +1,4 @@
 /* Copyright (c) 2003, David Leonard. All rights reserved. */
-/* $Id: parse.h 1215 2007-04-25 07:29:41Z d $ */
 
 #ifndef _SEE_h_parse_
 #define _SEE_h_parse_
@@ -23,7 +22,13 @@ void  SEE_functionbody_print(struct SEE_interpreter *i, struct function *f);
 struct SEE_string *SEE_functionbody_string(struct SEE_interpreter *i, 
 	struct function *f);
 int SEE_functionbody_isempty(struct SEE_interpreter *i, struct function *f);
-int SEE_compare(struct SEE_interpreter *i, struct SEE_value *x, 
-	struct SEE_value *y);
+
+void _SEE_call_eval(struct SEE_context *context, 
+        struct SEE_object *thisobj, int argc, struct SEE_value **argv, 
+        struct SEE_value *res);
+
+void _SEE_eval_input(struct SEE_context *context, 
+        struct SEE_object *thisobj, struct SEE_input *inp,
+        struct SEE_value *res);
 
 #endif /* _SEE_h_parse_ */
